import binascii
import os

from typing import Optional
from datetime import timedelta

from django.db import models
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser as BaseAbstractUser
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .typing import Context
from .email import Email
from .managers import UserManager

from .settings import settings


class AbstractUser(BaseAbstractUser):
    username_validator = None
    username = None

    email = models.EmailField(
        _('email address'),
        max_length=255,
        unique=True,
        error_messages={
            'unique': _('A user with that email address already exists.'),
        }
    )

    is_verified = models.BooleanField(
        _('verified'),
        default=False,
        help_text=_(
            'Designates whether this user has completed the email '
            'verification process to allow login.',
        ),
    )

    objects = UserManager()

    USERNAME_FIELD = 'email'
    REQUIRED_FIELDS = []

    class Meta:
        verbose_name = _('user')
        verbose_name_plural = _('users')
        abstract = True

    def __str__(self):
        return self.email


class AbstractCodeVerify(models.Model):
    """Docs."""

    code = models.CharField(
        _('code'),
        max_length=40,
        primary_key=True,
        editable=False,
    )
    user = models.ForeignKey(
        get_user_model(),
        on_delete=models.CASCADE,
        verbose_name=_('user'),
    )
    ipaddr = models.GenericIPAddressField(_('ip address'))
    link = models.URLField(
        _('email link'),
        blank=True, default='',
        help_text=_(
            'Contains a link to some source, will be supplemented with a '
            'request parameter with an instance code inside email.',
        )
    )

    created = models.DateTimeField(_('date created'), auto_now_add=True)

    email: Email | None = None
    expire_time: timedelta | None = timedelta(
        seconds=settings.USER_CODE_VERIFY_EXPIRE_TIME,
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if not self.code:
            self.code = self.generate_code()
        return super().save(*args, **kwargs)

    @classmethod
    def generate_code(cls):
        return binascii.hexlify(os.urandom(20)).decode()

    @classmethod
    def check_is_valid(cls, code: str, select_related_user: bool = False):
        if not isinstance(code, str) or not code:
            raise ValueError(
                _('The `code` attribute must be a non-empty string.'),
            )

        # Safe method to get instance by code
        queryset = cls.objects.filter(code=code)

        if select_related_user:
            queryset = queryset.select_related('user')

        instance = queryset.first()

        if instance is None:
            raise ValueError(_('The given `code` parameter is invalid.'))

        if instance.is_expired():
            raise ValueError(_('The given `code` parameter is expired.'))

        return instance

    def __str__(self):
        return self.code

    def _default_context(self) -> dict[str, str]:
        return {
            'email': self.user.email,
            'first_name': self.user.first_name,
            'last_name': self.user.last_name,
            'code': self.code,
            'link': self.link,
            'ipaddr': self.ipaddr,
        }

    def get_email(self) -> Email:
        assert self.email, (
            'The %s class `email` attribute was not provided.'
            % self.__class__.__name__,
        )
        return self.email

    def is_expired(self):
        if self.expire_time is None:
            return False

        return timezone.now() - self.created > self.expire_time

    def send_email(self, context: Optional[Context] = None):
        context = self._default_context() | ({} if context is None else context)
        self.get_email().send_email(target=self.user.email, context=context)



from typing import Any

from django.conf import settings as django_settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string


DEFAULTS = {
    'USER_EMAILS_CLASS': 'drf_auth_email.email.CodeVerifyEmail',
    'USER_EMAILS_WELCOME': 'drf_auth_email.emails.welcome_email',
    'USER_EMAILS_SIGNUP': 'drf_auth_email.emails.signup_email',
    'USER_EMAILS_PASSWORD_RESET': 'drf_auth_email.emails.password_reset_email',
    'USER_EMAILS_EMAIL_CHANGE_OLD_NOTIFY': 'drf_auth_email.emails.email_change_old_notify_email',
    'USER_EMAILS_EMAIL_CHANGE_NEW_CONFIRM': 'drf_auth_email.emails.email_change_new_confirm_email',

    'USER_EMAILS_DEFAULT_ORIGIN': 'http://127.0.0.1:8000',
    'USER_EMAIL_URL_BASENAME': 'drf-auth-email',

    # In seconds
    'USER_CODE_VERIFY_EXPIRE_TIME': 259200
}


# Settings with dot notation
IMPORT_STRINGS = [
    'USER_EMAILS_CLASS',
    'USER_EMAILS_WELCOME',
    'USER_EMAILS_SIGNUP',
    'USER_EMAILS_PASSWORD_RESET',
    'USER_EMAILS_EMAIL_CHANGE_OLD_NOTIFY',
    'USER_EMAILS_EMAIL_CHANGE_NEW_CONFIRM',
]


class CachedSettings:
    def __init__(self) -> None:
        self._cached = {}

    def __getattr__(self, __name: str) -> Any:
        if __name not in DEFAULTS:
            raise AttributeError('Invalid setting: %s' % __name)

        if __name in self._cached:
            return self._cached[__name]

        attr = getattr(django_settings, __name, DEFAULTS[__name])

        if __name in IMPORT_STRINGS:
            attr = import_string(attr)

        self._cached[__name] = attr
        return attr

    def reload(self):
        self._cached = {}


settings = CachedSettings()


def reload_settings(*args, **kwargs):
    settings.reload()


setting_changed.connect(reload_settings)



from typing import Optional

from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.conf import settings

from .typing import TemplateFiles, Context


def get_template_files(folder: str, prefix: str) -> TemplateFiles:
    """Get mail templates file paths by folder and prefix."""
    return TemplateFiles(
        '%s/%s_subject.txt' % (folder, prefix),
        '%s/%s.txt' % (folder, prefix),
        '%s/%s.html' % (folder, prefix),
    )


def send_multi_format_email(
    context: Context,
    target_email: str,
    folder: str = '',
    prefix: Optional[str] = None,
    templates: Optional[TemplateFiles] = None,
):
    _templates: TemplateFiles = None

    if templates is None and prefix is None:
        raise ValueError(
            'Can not to send multi-format email without template, specify '
            '`templates` or `prefix` attributes',
        )

    # Use `templates` attribute if this is not None
    if templates is not None:
        if isinstance(templates, TemplateFiles):
            _templates = templates

        # Necessary for some cases, for example when using Celery tasks, it
        # converts a 'TemplateFails' type variable into a regular `list`
        _templates = TemplateFiles(*templates)

    # Get templates using prefix and folder if `_templates` is None yet
    if _templates is None:
        _templates = get_template_files(folder, prefix)

    # Raise error if `_templates` variable can not be specified
    if _templates is None:
        # TODO: add folder, prefix and templates print
        raise ValueError(
            'Can not to specify templates to send multi-format email, check '
            'the given attributes.'
        )

    # Get attrs from settings
    from_email = settings.EMAIL_FROM
    to = target_email
    bcc_email = settings.EMAIL_BCC

    # Render templates with given context
    subject = render_to_string(_templates.subject).strip()
    text_content = render_to_string(_templates.txt, context)
    html_content = render_to_string(_templates.html, context)

    msg = EmailMultiAlternatives(
        subject, text_content,
        # Can be customized
        from_email, [to], bcc=[bcc_email],
    )
    msg.attach_alternative(html_content, 'text/html')
    msg.send()



from typing import Any

from django.db import models
from django.utils.translation import gettext_lazy as _
from django.conf import settings as django_settings

from .settings import settings
from .abstracts import AbstractCodeVerify


ABSTRACT = 'drf_auth_email' not in django_settings.INSTALLED_APPS


class SignupCode(AbstractCodeVerify):
    email = settings.USER_EMAILS_SIGNUP

    class Meta:
        verbose_name = _('signup code')
        verbose_name_plural = _('signup codes')
        abstract = ABSTRACT

    def verify_user(self) -> None:
        self.user.is_verified = True
        self.user.save()


class PasswordResetCode(AbstractCodeVerify):
    email = settings.USER_EMAILS_PASSWORD_RESET

    class Meta:
        verbose_name = _('password reset code')
        verbose_name_plural = _('password reset codes')
        abstract = ABSTRACT

    def change_user_password(self, password: str):
        self.user.set_password(password)
        self.user.save()


class EmailChangeCode(AbstractCodeVerify):
    new_email = models.EmailField(
        _('new email address'),
        max_length=255,
        help_text=_(
            'New email address will be related with user after email change '
            'verification.',
        )
    )

    class Meta:
        verbose_name = _('email change code')
        verbose_name_plural = _('email change codes')
        abstract = ABSTRACT

    def send_email(self, context: dict[str, Any] | None = None):
        context = (
            self._default_context()
            | ({} if context is None else context)
        )

        # send notification to old email address
        settings.USER_EMAILS_EMAIL_CHANGE_OLD_NOTIFY.send_email(
            target=self.user.email,
            context=context,
        )
        # send confirmation to new email address
        settings.USER_EMAILS_EMAIL_CHANGE_NEW_CONFIRM.send_email(
            target=self.new_email,
            context=context
        )

    def change_user_email(self):
        self.user.email = self.new_email
        self.user.save()



from django.contrib.auth.models import UserManager as BaseUserManager
from django.contrib.auth.hashers import make_password


class UserManager(BaseUserManager):
    def _create_user(
        self,
        email: str,
        password,
        **extra_fields,
    ):
        email = self.normalize_email(email)
        user = self.model(email=email, **extra_fields)
        user.password = make_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, email: str, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        extra_fields.setdefault("is_verified", False)

        return self._create_user(email, password, **extra_fields)

    def create_superuser(self, email: str, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)
        extra_fields.setdefault("is_verified", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")
        if extra_fields.get("is_verified") is not True:
            raise ValueError("Superuser must have is_verified=True.")

        return self._create_user(email, password, **extra_fields)



from .compat import OpenApiParameter, OpenApiExample, OpenApiResponse
from .serializers import DetailErrorSerializer


CodeQueryParameter = OpenApiParameter(
    name='code',
    type=str,
    location=OpenApiParameter.QUERY,
    description='Action code',
    required=True,
)
"""TODO: add docs."""


AuthorizationHeaderParameter = OpenApiParameter(
    name='authorization',
    type=str,
    location=OpenApiParameter.HEADER,
    description='Authorization token',
    required=True,
)
"""TODO: add docs."""


ErrorCodeResponse = OpenApiResponse(
    response=DetailErrorSerializer,
    description='Action code is invalid',
    examples=[
        OpenApiExample(
            'Code parameter is missing',
            description='string `code` parameter is missing',
            value={
                'detail': (
                    'The `code` attribute must be a non-empty '
                    'string.'
                ),
            },
        ),
        OpenApiExample(
            'Code parameter is invalid',
            description=(
                'could not find a model that related with '
                'provided code'
            ),
            value={
                'detail': 'The given `code` parameter is invalid.',
            },
        ),
        OpenApiExample(
            'Code is expired',
            description='code is expired',
            value={
                'detail': 'The given `code` parameter is expired.',
            },
        ),
    ],
)
"""TODO: add docs."""



# Check that settings user is drf_auth_email.user.abstracts.AbstractUser instance (for verified and unique email)
# Check that emails are drf_auth_email.user.email.CodeVerifyEmail instances
# Check that the user model has an required email field

# For views checks that user has an unique email field and boolean is_verified

# 'rest_framework' and 'rest_framework.authtoken' inside INSTALLED APPS

# check that all used templates exists



from rest_framework import serializers

from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password
from django.utils.translation import gettext_lazy as _


LINK_FIELD_HELP_TEXT = \
    'the link which can be send with the generated code inside email message'


class SignupSerializer(serializers.Serializer):
    """User signup serializer."""

    email = serializers.EmailField(max_length=255, help_text='the user email')
    password = serializers.CharField(
        max_length=128,
        help_text='the user password',
    )
    link = serializers.URLField(
        required=False,
        help_text=LINK_FIELD_HELP_TEXT,
    )

    first_name = serializers.CharField(
        max_length=150,
        required=False,
        help_text='the user first name',
    )
    last_name = serializers.CharField(
        max_length=150,
        required=False,
        help_text='the user last name',
    )


class PasswordResetSerializer(serializers.Serializer):
    """User password reset serializer."""

    email = serializers.EmailField(
        max_length=255,
        help_text='the user related email',
    )
    link = serializers.URLField(
        required=False,
        help_text=LINK_FIELD_HELP_TEXT,
    )


class PasswordResetVerifiedSerializer(serializers.Serializer):
    """Confirm password reset serializer."""

    password = serializers.CharField(max_length=128, help_text='new password')


class EmailChangeSerializer(serializers.Serializer):
    """Email change serializer."""

    email = serializers.EmailField(max_length=255, help_text='new email')
    link = serializers.URLField(
        required=False,
        help_text=LINK_FIELD_HELP_TEXT,
    )


class PasswordChangeSerializer(serializers.Serializer):
    """Password change serializer."""

    password = serializers.CharField(
        max_length=128,
        help_text='current password',
    )
    new_password = serializers.CharField(
        max_length=128,
        help_text='new password',
    )


class LoginSerializer(serializers.Serializer):
    """User login serializer."""

    email = serializers.EmailField(
        max_length=255,
        help_text='the user email',
    )
    password = serializers.CharField(
        max_length=128,
        help_text='the user password',
    )


class UserSerializer(serializers.ModelSerializer):
    """User model serializer.

    On `update()` or `create()` methods call generates hashed password from
    the provided password field.
    """

    class Meta:
        model = get_user_model()
        fields = ('id', 'email', 'password', 'first_name', 'last_name')
        extra_kwargs = {
            'password': {'write_only': True},
        }

    def update(self, instance, validated_data):
        validated_data['password'] = make_password(
            validated_data.get('password'),
        )
        return super().update(instance, validated_data)

    def create(self, validated_data):
        validated_data['password'] = make_password(
            validated_data.get('password'),
        )
        return super(UserSerializer, self).create(validated_data)


class DetailErrorSerializer(serializers.Serializer):
    """Detail error serializer.

    Used in OpenAPI schema only.
    """

    detail = serializers.CharField(help_text='error message')


class SuccessMessageSerializer(serializers.Serializer):
    """Success message serializer.

    Used in OpenAPI schema only.
    """
    success = serializers.CharField(help_text='success message')


class SuccessMessageWithEmailSerializer(serializers.Serializer):
    """Success message serializer.

    Contains additional `email` field. Used in OpenAPI schema only.
    """
    email = serializers.EmailField(help_text='success action related email')


class TokenSerializer(serializers.Serializer):
    """User authentication token serializer.

    Used in OpenAPI schema only.
    """
    token = serializers.CharField(help_text='user authentication token')
